#!/usr/bin/env python3
"""

Tracy: an experimental autotracer for glyphs

Copyright 2014 Damian Yerrick
Copying and distribution of this file, with or without modification,
are permitted in any medium without royalty provided the copyright
notice and this notice are preserved.  This file is offered as-is,
without any warranty.

"""
from __future__ import with_statement, division, print_function
from PIL import Image, ImageDraw
import array
from itertools import chain
try:
    from html import escape as H
except ImportError:
    from cgi import escape as cgi_escape
    def H(s):
        return cgi_escape(s, True)

direction_names = 'rdlu'
directions = [(1, 0), (0, 1), (-1, 0), (0, -1)]

def find_all_walls(im, bgcolor):
    """Find all "walls" (transitions to and from bgcolor) in an image."""
    xgrad = set()
    ygrad = set()
    w, h = im.size
    px = im.load()
    for y in range(h):
        rp = bgcolor
        for x in range(w):
            lp, rp = rp, px[x, y]
            if rp != lp:
                xgrad.add((x, y, rp - lp))
        if bgcolor != rp:
            xgrad.add((w, y, bgcolor - rp))
    for x in range(w):
        rp = bgcolor
        for y in range(h):
            lp, rp = rp, px[x, y]
            if rp != lp:
                ygrad.add((x, y, rp - lp))
        if bgcolor != rp:
            ygrad.add((x, h, bgcolor - rp))

    # An xgrad (32, 8, 1) means that between the pixels (31, 8) and
    # (32, 8), the height increases by 1.  This corresponds to a
    # line segment between the points (32, 8) and (32, 9).
    return xgrad, ygrad


def trace_all_walls(glyph):
    """Trace the contours in a bilevel indexed glyph.

Contours are placed between pixels.  The left side is the outside,
and the right side is the inside.  This means outer contours are
clockwise, inner ones counterclockwise.
Each starts at the top of the leftmost extremum, meaning that outer
contours will start with r and end with (usually several) u, and
inner contours will begin with (usually several) d and end with l.

Return [((startx, starty), [dir, ...]), ...]
where (startx, starty) are the point coordinates of the start of a
contour, and dir is a sequence of 0 for right, 1 for down, 2 for
left, and 3 for up.

"""
    xgrad, ygrad = find_all_walls(glyph, 0)
    contours = []
    while xgrad and ygrad:
##        print("Walls left:", len(xgrad) + len(ygrad))
        startpos = min(xgrad)[:2]
        x, y = startpos
        foundpath = array.array('B')
        while not foundpath or (x, y) != startpos:
            if (x, y, 1) in ygrad:  # Try to the right
                ygrad.remove((x, y, 1))
                foundpath.append(0)
                x += 1
            elif (x, y, -1) in xgrad:  # Try down
                xgrad.remove((x, y, -1))
                foundpath.append(1)
                y += 1
            elif (x - 1, y, -1) in ygrad:  # Try left
                ygrad.remove((x - 1, y, -1))
                foundpath.append(2)
                x -= 1
            elif (x, y - 1, 1) in xgrad:  # Try up
                xgrad.remove((x, y - 1, 1))
                foundpath.append(3)
                y -= 1
            else:
                print("""internal error: started at (%d, %d), made it to (%d, %d)
path so far is %s"""
                      % (startpos[0], startpos[1], x, y,
                         ''.join(direction_names[x] for x in foundpath)))
                return None
        contours.append((startpos, foundpath))
    return contours

def reporting1(glyphs):
    tpaths = []
    print("reporting")
    for i, g in enumerate(glyphs):
        tpaths.append("== Glyph $%02X ==" % (i + 32))
        tpaths.extend("Start at (%d, %d) and go\n%s"
                      % (startpos[0], startpos[1],
                         ''.join(direction_names[x] for x in foundpath))
                      for startpos, foundpath in g)
    print("\n".join(tpaths))

def contour_to_polyline(contour):
    """Iterate through the points in a contour."""
    ((x, y), dirs) = contour
    out = []
    for d in dirs:
        yield (x, y)
        dx, dy = directions[d]
        x += dx
        y += dy

def local_maxima(data, tolerance=0):
    """Find local maxima of a circular list.

data -- a list of coordinates
tolerance -- size of dips permitted in polyline
"""
    found_maxima = []

    # Find global minimum and start from it, so that maxima on the
    # circular list don't get split
    hi_ht, hi_i = min((ht, i) for (i, ht) in enumerate(data))
    indices = chain(range(hi_i, len(data)), range(hi_i))
    for i in indices:
        ht = data[i]
##        print("i =%3d  ht =%3d  hi_ht =%3d  hi_i %s"
##              % (i, ht, hi_ht, "Y" if hi_i is not None else ' '))
        if ht > hi_ht and (hi_i is not None or ht > hi_ht + tolerance):
            hi_i, hiend_i, hi_ht = i, i, ht
        elif ht >= hi_ht and hi_i is not None:
            hiend_i = i
        elif ht < hi_ht - tolerance:
            if hi_i is not None:
                found_maxima.append((hi_i, hiend_i + 1))
##                print("wrote %s" % repr(found_maxima[-1]))
                hi_i = None
            hi_ht = ht
    if hi_i is not None:
        # In case the last local maximum occurs immediately before
        # the global minimum
        found_maxima.append((hi_i, len(data)))
    return found_maxima

def polyline_dotproduct(polyline, direction):
    """Calculate the dot product of each coordinate with a vector.

polyline -- a list of coordinates as 2-tuples
direction -- a 2-tuple representing a vector
"""
    return [direction[0] * x + direction[1] * y for (x, y) in polyline]

# The eight directions in which to try to insert parabola endpoints
maxima_dirs = [
    (1, 0), (1, 1), (0, 1), (-1, 1), (-1, 0), (-1, -1), (0, -1), (1, -1)
]
maxima_dirnames = [
    'e', 'se', 's', 'sw', 'w', 'nw', 'n', 'ne'
]

def find_contour_maxima(polyline):
    """Find the local maxima of a polyline.

Return them as a list of 4-tuples (order, start, end, heading) where
order: approximate index of midpoint of maximum
start, end: indices of endpoints of maximum
heading: which direction is normal to the vector

Start your curve fitter by placing on-curve points halfway between
start and end and off-curve points at intersections of tangent
lines represented by heading.  Use small start-end gaps and small
differences in order between consecutive maxima to detect corners.
Use large differences in heading to detect inflection points, or
places where the contour switches between convex and concave.

"""

    hts = [(polyline_dotproduct(polyline, direction),
            1 if heading % 2 else 0)
           for heading, direction in enumerate(maxima_dirs)]
    slices = [(local_maxima(dp, tolerance)) for (dp, tolerance) in hts]
    exslices = [((start + end + (len(polyline) if end < start else 0)) // 2,
                 start, end, heading)
                for heading, mx in enumerate(slices) for start, end in mx]
    exslices.sort()
    return exslices

def add_inflection_points(polyline, exslices):
    it = iter(exslices)
    _, _, preve, prevnorm = firstpt = next(it)
    out = [None]  # reserve first element
    for _, s, e, norm in chain(it, [firstpt]):
        normdiff = (prevnorm - norm + 2) % 8
        sx, sy = polyline[s]
        ex, ey = polyline[e - 1]
        x = (sx + ex) / 2
        y = (sy + ey) / 2
        if normdiff > 4:
            prevex, prevey = polyline[preve - 1]
            out.append(((sx + prevex) / 2, (sy + prevey) / 2, 'i', (prevex, prevey)))
        out.append((x, y, norm))
        prevnorm, preve, prevx, prevy = norm, e, x, y
    out[0] = out.pop()
    return out

def solve_3x2_system(ex0, ey0, ers0, ex1, ey1, ers1):
    if ((ex1 and not ex0)
        or (ex1 and ex0 and not ey0)):
        ex0, ex1, ey0, ey1, ers0, ers1 = ex1, ex0, ey1, ey0, ers1, ers0
    assert ex0
##    if ex0 == 0:
##        print("solve this:\n%2dx + %2dy = %d\n%2dx + %2dy = %d"
##              % (ex0, ey0, ers0, ex1, ey1, ers1))
    # Get ex0 = 1
    if ex0 not in (0, 1):
        fac = 1/ex0
        ex0, ey0, ers0 = 1, ey0*fac, ers0*fac
    assert ex0 == 1
    # Get ex1 = 0 (completing the row echelon)
    if ex0 and ex1:
        fac = ex1
        ex1, ey1, ers1 = 0, ey1 - ey0*fac, ers1 - ers0*fac
    assert ex1 == 0
    assert ey1 != 0
    # Get ey1 = 1
    if ey1 != 1:
        fac = 1/ey1
        ey1, ers1 = 1, ers1*fac
    # Get ey0 = 0 (completing the reduced row echelon)
    if ey0 != 0:
        fac = ey0
        ey0, ers0 = 0, ers0 - ers1*fac
    return ers0, ers1

def add_offcurve_points(oncurvepts):
    """Place offcurve points at intersections of oncurve points' tangent lines."""
    it = iter(oncurvepts)
    x0, y0, norm0 = firstpt = next(it)
    out = []
    inflparity = 1

    # solve two simutaneous equations of the following form:
    # normx * x + normy * y = (normx * x0 + normy * y0)
    normx0, normy0 = maxima_dirs[norm0]
    normrs0 = (normx0 * x0 + normy0 * y0)
    for row in chain(it, [firstpt]):
        x1, y1, norm1 = row[:3]
        if norm1 == 'i':

            # Inflection point vertex is at the midpoint between
            # ends of extrema.  Its control points are partway
            # between the extremum end and the perpendicular
            # projection onto the extremum's tangent.  Find
            # this projection first.
            endx, endy = row[3]
            normrs1 = (normy0 * x1 - normx0 * y1)
            crosspt = solve_3x2_system(normx0, normy0, normrs0,
                                       normy0, -normx0, normrs1)
            cpx = (crosspt[0] + 2 * endx) / 3
            cpy = (crosspt[1] + 2 * endy) / 3

            # Now compute the inflection point's normal vector
            # to allow calculation of the off-curve point on the
            # other side of the inflection point
            normx1 = (cpy - y1)
            normy1 = (x1 - cpx)
        else:
            normx1, normy1 = maxima_dirs[norm1]
        normrs1 = (normx1 * x1 + normy1 * y1)

        if normx0 * normy1 == normx1 * normy0:
##            print("Zero cross product between vertex normals; parallel?")
            out.append(((x0 + x1) / 2, (y0 + y1)/2))
        else:
            cpx, cpy = solve_3x2_system(normx0, normy0, normrs0, normx1, normy1, normrs1)
            out.append((cpx, cpy))

        x0, y0, norm0 = x1, y1, norm1
        normx0, normy0, normrs0 = normx1, normy1, normrs1
    return out

def quadbezier_to_polyline(start, mid, end):
    x0, y0 = start
    xm, ym = mid
    x4, y4 = end
    x1, y1 = (x0 + xm) / 2, (y0 + ym) / 2
    x3, y3 = (x4 + xm) / 2, (y4 + ym) / 2
    x2, y2 = (x1 + x3) / 2, (y1 + y3) / 2
    
    if abs(xm - x2) < 2 and abs(ym - y2) < 2:
        return [(x0, y0), (x2, y2), (x4, y4)]
    lhalf = quadbezier_to_polyline(start, (x1, y1), (x2, y2))
    lhalf.pop()
    lhalf.extend(quadbezier_to_polyline((x2, y2), (x3, y3), end))
    return lhalf

# PRESENTATION ######################################################

def contour_to_svg(onc, offc):
    onci = chain(onc, [onc[0]])
    cx, cy = next(onci)[:2]
    words = ['M%.1f %.1f' % (cx, cy)]
    for (hx, hy), oncp in zip(offc, onci):
        (cx, cy) = oncp[:2]
        words.append('Q%.1f %.1f %.1f %.1f' % (hx, hy, cx, cy))
    words.append('Z')
    return ' '.join(words)

def uv2rgb(u, v):
    """Create an RGB triplet from UV in (-25...25) range."""
    # based on matrix from http://stackoverflow.com/q/9325861/2738262
    return (128 + 4 * v, 128 - 2 * v - u, 128 + 5 * u)

def disp_contours(contours, filename=None):
    scale = 8
    margin = 2
    linewid = 1
    circlerad = 5
    offcrad = 3
    polylines = [list(contour_to_polyline(c)) for c in contours]
    xmin = min(pt[0] for c in polylines for pt in c)
    xmax = max(pt[0] for c in polylines for pt in c)
    ymin = min(pt[1] for c in polylines for pt in c)
    ymax = max(pt[1] for c in polylines for pt in c)
    imsz = ((xmax - xmin + 2 * margin) * scale, (ymax - ymin + 2 * margin) * scale)
    im = Image.new('RGB', imsz, (0, 0, 0))
    dc = ImageDraw.Draw(im)

    circles = []
    bigtrace = []

    # Collect control points and draw the line connecting them
    colors = [(direction[0], direction[1], 17 if i % 2 else 25)
              for i, direction in enumerate(maxima_dirs)]
    colors = [uv2rgb(-v * f, u * f) for (u, v, f) in colors]
    hullcolor = (112, 112, 112)
    for c in polylines:
        bigtrace.append([])
        maxima = find_contour_maxima(c)
        onc = add_inflection_points(c, maxima)
        offc = add_offcurve_points(onc)
        prevhx, prevhy = offc[-1][:2]
        prevcx, prevcy = onc[-1][:2]
        prevcx = int(round((prevcx - xmin + 2) * scale))
        prevcy = int(round((prevcy - ymin + 2) * scale))
        prevhx = int(round((prevhx - xmin + 2) * scale))
        prevhy = int(round((prevhy - ymin + 2) * scale))
        for (hx, hy), oncp in zip(offc, onc):
            (cx, cy, heading) = oncp[:3]
            hx = int(round((hx - xmin + 2) * scale))
            hy = int(round((hy - ymin + 2) * scale))
            cx = int(round((cx - xmin + 2) * scale))
            cy = int(round((cy - ymin + 2) * scale))
            engon = quadbezier_to_polyline((prevcx, prevcy), (prevhx, prevhy), (cx, cy))
            bigtrace[-1].extend(engon)
            arcbox = (hx - offcrad, hy - offcrad, hx + offcrad, hy + offcrad)
            circles.append(([int(round(p)) for p in arcbox],
                            0, 360, None, hullcolor))
            dc.line((hx, hy, cx, cy), hullcolor)
            dc.line((cx, cy, prevhx, prevhy), hullcolor)

            arcbox = (cx - circlerad, cy - circlerad, cx + circlerad, cy + circlerad)
            if heading == 'i':
                start, end = 0, 360
                fillcolor = hullcolor
            else:
                start, end = (heading - 2) * 45, (heading + 2) * 45
                fillcolor = colors[heading]
            circles.append(([int(round(p)) for p in arcbox],
                            start, end, None, fillcolor))
            prevcx, prevcy, prevhx, prevhy = cx, cy, hx, hy
        dc.polygon(bigtrace[-1], outline=(192, 192, 192))

    # Draw the actual shape's contour
    for c in polylines:
        prevpt = c[-1]
        for i, pt in enumerate(c):
            tana = c[i - 3]
            tanb = c[i - 2]
            tanc = c[(i + 1) % len(c)]
            tand = c[(i + 2) % len(c)]
            tanx = (tanc[0] + tand[0] - tana[0] - tanb[0]) * 3
            tany = (tanc[1] + tand[1] - tana[1] - tanb[1]) * 3
            # Modified YUV decoding matrix
            fillcolor = uv2rgb(tanx, tany)

            coords = (((prevpt[0] - xmin + margin) * scale,
                       (prevpt[1] - ymin + margin) * scale),
                      ((pt[0] - xmin + margin) * scale,
                       (pt[1] - ymin + margin) * scale))
            dc.line(coords, fill=fillcolor, width=linewid)
            prevpt = pt

    # Draw the control points
    for c in circles:
        dc.chord(*c)

    return im

# The "contours" are closed paths as described at
# http://pkgs.fedoraproject.org/lookaside/pkgs/potrace/potrace.pdf/a58c3d99e0213daa0a9f7d1cd253f75e/potrace.pdf
# Unlike more general tools such as potrace, this has a distinct
# philosophy on how to convert the outline to parabolic arcs,
# inspired by the chapter "Digitizing letterform designs" in the
# TrueType Reference Manual by Apple Inc.
# https://developer.apple.com/fonts/ttrefman/RM01/Chap1.html
# Each 45 degrees of curvature gets its own arc, so find local
# maxima in eight directions and put vertices there.   A vertex
# also goes at inflection points of the outline, which occurs
# somewhere between consecutive vertices with opposite normals.
#
# I'll also need to put nodes at corners, but I'll need to wait
# for inspiration on means to identify those.

im = Image.open("font64.png")
w, h = im.size
glyphs = []
gw, gh = 64, 64
print("tracing")
for y in range(0, h, gh):
    for x in range(0, w, gw):
        zero = im.crop((x, y, x + gw, y + gh))
        glyphs.append(trace_all_walls(zero))
print("all traced")

##zero = glyphs[ord('4')-32]
##disp_contours(zero).show()

# Now try generating an SVG file of all glyphs.
# A lot of defective SVG renderers require the prefix for the XLink
# namespace to be exactly "xlink", not the more compact "L".
# These include at least EOG Image Viewer (included with GNOME)
# and ImageMagick (used by MediaWiki).
svgfile = ["""<svg version="1.1"
 xmlns="http://www.w3.org/2000/svg"
 xmlns:xlink="http://www.w3.org/1999/xlink"
 width='512' height='512'>
<!-- Generated with tracy, an experimental glyph tracer -->
<defs>
"""]
svguses = []

for i, contours in enumerate(glyphs):
    polylines = [list(contour_to_polyline(c)) for c in contours]
    if not polylines:  # space character
        continue
    pathdata = []
    for c in polylines:
        maxima = find_contour_maxima(c)
        onc = add_inflection_points(c, maxima)
        offc = add_offcurve_points(onc)
        pathdata.append(contour_to_svg(onc, offc))
    svgfile.append('<g id="u%04x"><path d="%s" /></g>\n'
                   % (i + 32, H(' '.join(pathdata))))
    svguses.append('<use xlink:href="#u%04x" x="%d" y="%d" />\n'
                   % (i + 32, (i % 8) * 64, (i // 8) * 64))
svgfile.append("""</defs>
<g fill="#555555" fill-opacity="1.0" stroke="#000000" stroke-width="1px">
""")
svgfile.append(''.join(svguses))
svgfile.append("</g></svg>")
svgfile = ''.join(svgfile)

if str is not bytes:
    svgfile = svgfile.encode('ascii')    
with open("tracyout.svg", "wb") as outfp:
    outfp.write(svgfile)
